import pandas as pd
import plotly.graph_objects as go

from .format_tools import plotly_figure_to_html_plot
from .plotly_bias_corrected import (
    corrected_retrospective as plotly_corrected_retrospective,
    corrected_month_average as plotly_corrected_month_average,
    corrected_day_average as plotly_corrected_day_average,
    corrected_scatterplots as plotly_corrected_scatterplots,
)
from .plotly_forecasts import (
    forecast as plotly_forecast,
    forecast_stats as plotly_forecast_stats,
    forecast_ensembles as plotly_forecast_ensembles,
    forecast_records as plotly_forecast_records,
)
from .plotly_retrospective import (
    retrospective as plotly_retrospective,
    daily_averages as plotly_daily_averages,
    monthly_averages as plotly_monthly_averages,
    annual_averages as plotly_annual_averages,
    flow_duration_curve as plotly_flow_duration_curve,
)

__all__ = [
    'forecast',
    'forecast_stats',
    'forecast_ensembles',
    'forecast_records',

    'retrospective',
    'daily_averages',
    'monthly_averages',
    'annual_averages',
    'flow_duration_curve',

    'corrected_retrospective',
    'corrected_month_average',
    'corrected_day_average',
    'corrected_scatterplots',
]


def forecast(df: pd.DataFrame, *,
             plot_type: str = 'plotly',
             rp_df: pd.DataFrame = None,
             plot_titles: list = None, ) -> go.Figure:
    """
    Plots forecasted streamflow and optional return periods
    Args:
        df: the dataframe response from geoglows.data.forecast
        rp_df: optional dataframe of return period data
        plot_titles: optional list of strings to include in the figure title. each list item will be on a new line.

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_forecast(df, rp_df=rp_df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def forecast_stats(df: pd.DataFrame, *,
                   plot_type: str = 'plotly',
                   rp_df: pd.DataFrame = None,
                   plot_titles: list = None, ) -> go.Figure:
    """
    Plots forecasted streamflow and optional return periods
    Args:
         df: the dataframe response from geoglows.data.forecast_stats
         rp_df: optional dataframe of return period data
         plot_titles: optional list of strings to include in the figure title. each list item will be on a new line.

    Returns:
         go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_forecast_stats(df, rp_df=rp_df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def forecast_ensembles(df: pd.DataFrame, *,
                       plot_type: str = 'plotly',
                       rp_df: pd.DataFrame = None,
                       plot_titles: list = None, ) -> go.Figure:
    """
    Plots forecasted streamflow and optional return periods
    Args:
           df: the dataframe response from geoglows.data.forecast_ensembles
           rp_df: optional dataframe of return period data
           plot_titles: optional list of strings to include in the figure title. each list item will be on a new line.

    Returns:
           go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_forecast_ensembles(df, rp_df=rp_df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def forecast_records(df: pd.DataFrame, *,
                     plot_type: str = 'plotly',
                     rp_df: pd.DataFrame = None,
                     plot_titles: list = None, ) -> go.Figure:
    """
    Plots forecasted streamflow and optional return periods
    Args:
        df:
        plot_type:
        rp_df:
        plot_titles:

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_forecast_records(df, rp_df=rp_df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def retrospective(df: pd.DataFrame, *,
                  plot_type: str = 'plotly',
                  rp_df: pd.DataFrame = None,
                  plot_titles: list = None, ) -> go.Figure:
    """
    Plots the retrospective simulation data

    Args:
        df: a dataframe of the retrospective simulation
        plot_type: either plotly or html
        rp_df: a dataframe with the response from the return periods dataset
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_retrospective(df, rp_df=rp_df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def daily_averages(df: pd.DataFrame, *,
                   plot_type: str = 'plotly',
                   plot_titles: list = None, ) -> go.Figure:
    """
    Makes a plotly figure of the daily average flows

    Args:
        df: a dataframe of the daily average flows
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_daily_averages(df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def monthly_averages(df: pd.DataFrame, *,
                     plot_type: str = 'plotly',
                     plot_titles: list = None, ) -> go.Figure:
    """
    Makes a plotly figure of the monthly average flows

    Args:
        df: a dataframe of the monthly average flows
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_monthly_averages(df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def annual_averages(df: pd.DataFrame, *,
                    plot_type: str = 'plotly',
                    plot_titles: list = None,
                    decade_averages: bool = False, ) -> go.Figure:
    """
    Makes a plotly figure of the annual average flows

    Args:
        df: a dataframe of the annual average flows
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure
        decade_averages: if True, the figure will include the average flows for each decade

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_annual_averages(df, plot_titles=plot_titles, decade_averages=decade_averages)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def flow_duration_curve(df: pd.DataFrame, *,
                        plot_type: str = 'plotly',
                        plot_titles: list = None, ) -> go.Figure:
    """
    Makes a plotly figure of the flow duration curve

    Args:
        df: a dataframe of the flow duration curve values
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_flow_duration_curve(df, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def corrected_retrospective(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                            rperiods: pd.DataFrame = None, plot_titles: list = None,
                            plot_type: str = 'plotly', ) -> go.Figure:
    """
    Makes a plotly figure of bias corrected retrospective simulation data

    Args:
        corrected: the bias corrected simulation dataframe
        simulated: the retrospective simulation dataframe
        observed: the in situ measurements used to bias correct the simulation
        rperiods: the return periods for the simulated river
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_corrected_retrospective(corrected, simulated, observed, rperiods, plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    else:
        raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def corrected_month_average(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                            plot_titles: list = None, plot_type: str = 'plotly', ) -> go.Figure:
    """
    Makes a plotly figure of bias corrected monthly average simulation data

    Args:
        corrected: the bias corrected simulation dataframe
        simulated: the retrospective simulation dataframe
        observed: the in situ measurements used to bias correct the simulation
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_corrected_month_average(corrected, simulated, observed, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    else:
        raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def corrected_day_average(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                          plot_titles: list = None, plot_type: str = 'plotly', ) -> go.Figure:
    """
    Makes a plotly figure of bias corrected daily average simulation data

    Args:
        corrected: the bias corrected simulation dataframe
        simulated: the retrospective simulation dataframe
        observed: the in situ measurements used to bias correct the simulation
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_corrected_day_average(corrected, simulated, observed, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    else:
        raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')


def corrected_scatterplots(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                           plot_titles: list = None, plot_type: str = 'plotly', ) -> go.Figure:
    """
    Makes a plotly figure of scatter _plots showing the performance of the bias correction process

    Args:
        corrected: the bias corrected simulation dataframe
        simulated: the retrospective simulation dataframe
        observed: the in situ measurements used to bias correct the simulation
        plot_type: either plotly or html
        plot_titles: additional key-value pairs to display in the title of the figure

    Returns:
        go.Figure
    """
    if plot_type in ('plotly', 'html'):
        figure = plotly_corrected_scatterplots(corrected, simulated, observed, plot_titles=plot_titles)
        if plot_type == 'html':
            return plotly_figure_to_html_plot(figure)
        return figure
    else:
        raise NotImplementedError(f'Plot type "{plot_type}" is not supported.')



import numpy as np

from .data import metadata_tables

__all__ = ['river_to_vpu', 'latlon_to_river', 'river_to_latlon', ]


def river_to_vpu(river_id: int, metadata_table_path: str = None) -> int:
    """
    Gives the VPU number for a given River ID number

    Args:
        river_id (int): a 9 digit integer that is a valid GEOGLOWS River ID number
        metadata_table_path (str): optional path to the local metadata table

    Returns:
        int: a 3 digit integer that is the VPU number for the given River ID number
    """
    return (
        metadata_tables(columns=['LINKNO', 'VPUCode'], metadata_table_path=metadata_table_path)
        .loc[lambda x: x['LINKNO'] == river_id, 'VPUCode']
        .values[0]
    )


def latlon_to_river(lat: float, lon: float, metadata_table_path: str = None) -> int:
    """
    Gives the River ID number whose outlet is nearest the given lat and lon
    Args:
        lat (float): a latitude
        lon (float): a longitude
        metadata_table_path (str): optional path to the local metadata table

    Returns:
        int: a 9 digit integer that is a valid GEOGLOWS River ID number
    """
    df = metadata_tables(columns=['LINKNO', 'lat', 'lon'], metadata_table_path=metadata_table_path)
    df['dist'] = ((df['lat'] - lat) ** 2 + (df['lon'] - lon) ** 2) ** 0.5
    return df.loc[lambda x: x['dist'] == df['dist'].min(), 'LINKNO'].values[0]


def river_to_latlon(river_id: int, metadata_table_path: str = None) -> np.ndarray:
    """
    Gives the lat and lon of the outlet of the river with the given River ID number

    Args:
        river_id (int): a 9 digit integer that is a valid GEOGLOWS River ID number
        metadata_table_path (str): optional path to the local metadata table

    Returns:
        np.ndarray: a numpy array of floats, [lat, lon]
    """
    return (
        metadata_tables(columns=['LINKNO', 'lat', 'lon'], metadata_table_path=metadata_table_path)
        .loc[lambda x: x['LINKNO'] == river_id, ['lat', 'lon']]
        .values[0]
    )



import math
import warnings

import hydrostats as hs
import hydrostats.data as hd
import numpy as np
import pandas as pd
from scipy import interpolate
from scipy import stats

from .data import sfdc_for_river_id
from .data import retrospective

__all__ = [
    'correct_historical',
    'correct_forecast',
    'statistics_tables',
]


def correct_historical(simulated_data: pd.DataFrame, observed_data: pd.DataFrame) -> pd.DataFrame:
    """
    Accepts a historically simulated flow timeseries and observed flow timeseries and attempts to correct biases in the
    simulation on a monthly basis.

    Args:
        simulated_data: A dataframe with a datetime index and a single column of streamflow values
        observed_data: A dataframe with a datetime index and a single column of streamflow values

    Returns:
        pandas DataFrame with a datetime index and a single column of streamflow values
    """
    # list of the unique months in the historical simulation. should always be 1->12 but just in case...
    unique_simulation_months = sorted(set(simulated_data.index.strftime('%m')))
    dates = []
    values = []

    for month in unique_simulation_months:
        # filter historic data to only be current month
        monthly_simulated = simulated_data[simulated_data.index.month == int(month)].dropna()
        to_prob = _flow_and_probability_mapper(monthly_simulated, to_probability=True)
        # filter the observations to current month
        monthly_observed = observed_data[observed_data.index.month == int(month)].dropna()
        to_flow = _flow_and_probability_mapper(monthly_observed, to_flow=True)

        dates += monthly_simulated.index.to_list()
        value = to_flow(to_prob(monthly_simulated.values))
        values += value.tolist()

    corrected = pd.DataFrame(data=values, index=dates, columns=['Corrected Simulated Streamflow'])
    corrected.sort_index(inplace=True)
    return corrected


def sfdc_bias_correction(river_id: int) -> pd.DataFrame:
    """
    Corrects the bias in the simulated data for a given river_id using the SFDC method. This method is based on the
     https://github.com/rileyhales/saber-hbc repository.

    Args: river_id: int: a 9 digit integer that is a valid GEOGLOWS River ID number

    Returns: pandas DataFrame with a DateTime index and columns with Simulated flow, Bias Corrected Simulation flow.

    """
    simulated_data = retrospective(river_id=river_id)
    print(simulated_data)
    sfdc_b = sfdc_for_river_id(river_id=river_id)
    sim_fdc_b = []
    for i in range(1, 13):
        mdf = fdc(simulated_data[simulated_data.index.month == i].values.flatten()).rename(columns={'Q': 'fdc'}).reset_index()
        mdf['month'] = i
        sim_fdc_b.append(mdf)
    sim_fdc_b = pd.concat(sim_fdc_b, axis=0)

    # Process each month from January (1) to December (12)
    monthly_results = []
    for month in sorted(set(simulated_data.index.month)):
        mon_sim_b = simulated_data[simulated_data.index.month == month].dropna().clip(lower=0)
        qb_original = mon_sim_b[river_id].values.flatten()
        scalar_fdc = sfdc_b[sfdc_b['month'] == month][['p_exceed', 'sfdc']].set_index('p_exceed')
        sim_fdc_b_m = sim_fdc_b[sim_fdc_b['month'] == month][['p_exceed', 'fdc']].set_index('p_exceed')

        flow_to_percent = _make_interpolator(sim_fdc_b_m.values.flatten(),
                                             sim_fdc_b_m.index,
                                             extrap='nearest',
                                             fill_value=None)

        percent_to_scalar = _make_interpolator(scalar_fdc.index,
                                               scalar_fdc.values.flatten(),
                                               extrap='nearest',
                                               fill_value=None)
        p_exceed = flow_to_percent(qb_original)
        scalars = percent_to_scalar(p_exceed)
        qb_adjusted = qb_original / scalars

        # Create a DataFrame for this month's results
        month_df = pd.DataFrame({
            'date': mon_sim_b.index,
            'Simulated': qb_original,
            'Bias Corrected Simulation': qb_adjusted
        })

        # Append the month's DataFrame to the results list
        monthly_results.append(month_df)
    return pd.concat(monthly_results).set_index('date').sort_index()


def make_sfdc(simulated_df: pd.DataFrame, gauge_df: pd.DataFrame,
              use_log: bool = False, fix_seasonally: bool = True, empty_months: str = 'skip',
              drop_outliers: bool = False, outlier_threshold: int or float = 2.5,
              filter_scalar_fdc: bool = False, filter_range: tuple = (0, 80),
              extrapolate: str = 'nearest', fill_value: int or float = None,
              fit_gumbel: bool = False, fit_range: tuple = (10, 90),
              metadata: bool = False, ) -> pd.DataFrame:
    """
    Makes a scalar flow duration curve (SFDC) mapping from simulated to observed (gauge) flow data. SFDC is used in SABER to
    correct timeseries data for a ungauged riverid.

    Args:
        simulated_data: A dataframe with a datetime index and a single column of simulated streamflow values
        gauge_df: A dataframe with a datetime index and a single column of observed streamflow values
        extrapolate: Method to use for extrapolation. Options: nearest, const, linear, average, max, min
        fill_value: Value to use for extrapolation when extrapolate_method='const'
        filter_range: Lower and upper bounds of the filter range
        drop_outliers: Flag to exclude outliers
        outlier_threshold: Number of std deviations from mean to exclude from flow duration curve

    Returns:
        pandas DataFrame with a DateTime index and columns with corrected flow, uncorrected flow, the scalar adjustment
        factor applied to correct the discharge, and the percentile of the uncorrected flow (in the seasonal grouping,
        if applicable).
    """
    if fix_seasonally:
        # list of the unique months in the historical simulation. should always be 1->12 but just in case...
        monthly_results = []
        for month in sorted(set(simulated_df.index.strftime('%m'))):
            # filter data to current iteration's month
            mon_obs_a = gauge_df[gauge_df.index.month == int(month)].dropna().clip(lower=0)
            if mon_obs_a.empty:
                if empty_months == 'skip':
                    continue
                else:
                    raise ValueError(f'Invalid value for argument "empty_months". Given: {empty_months}.')

            mon_sim_b = simulated_df[simulated_df.index.month == int(month)].dropna().clip(lower=0)
            monthly_results.append(make_sfdc(
                mon_obs_a, mon_sim_b,
                fix_seasonally=False, empty_months=empty_months,
                drop_outliers=drop_outliers, outlier_threshold=outlier_threshold,
                filter_scalar_fdc=filter_scalar_fdc, filter_range=filter_range,
                extrapolate=extrapolate, fill_value=fill_value,
                fit_gumbel=fit_gumbel, fit_range=fit_range, )
            )
        # combine the results from each monthly into a single dataframe (sorted chronologically) and return it
        return pd.concat(monthly_results).sort_index()

    if drop_outliers:
        simulated_df = _drop_outliers_by_zscore(simulated_df, threshold=outlier_threshold)
        gauge_df = _drop_outliers_by_zscore(gauge_df, threshold=outlier_threshold)

    # compute the flow duration curves
    sim_fdc = fdc(simulated_df)
    obs_fdc = fdc(gauge_df)

    # calculate the scalar flow duration curve
    scalar_fdc = sfdc(sim_fdc, obs_fdc)

    return scalar_fdc


def fdc(flows: np.array, steps: int = 101, col_name: str = 'Q') -> pd.DataFrame:
    """
    Compute flow duration curve (exceedance probabilities) from a list of flows

    Args:
        flows: array of flows
        steps: number of steps (exceedance probabilities) to use in the FDC
        col_name: name of the column in the returned dataframe

    Returns:
        pd.DataFrame with index 'p_exceed' and columns 'Q' (or col_name)
    """
    # calculate the FDC and save to parquet
    exceed_prob = np.linspace(100, 0, steps)
    fdc_flows = np.nanpercentile(flows, exceed_prob)
    df = pd.DataFrame(fdc_flows, columns=[col_name, ], index=exceed_prob)
    df.index.name = 'p_exceed'
    return df


def _drop_outliers_by_zscore(df: pd.DataFrame, threshold: float = 3) -> pd.DataFrame:
    """
    Drop outliers from a dataframe by their z-score and a threshold
    Based on https://stackoverflow.com/questions/23199796/detect-and-exclude-outliers-in-pandas-data-frame
    Args:
        df: dataframe to drop outliers from
        threshold: z-score threshold

    Returns:
        pd.DataFrame with outliers removed
    """
    return df[(np.abs(stats.zscore(df)) < threshold).all(axis=1)]


def sfdc(sim_fdc: pd.DataFrame, obs_fdc: pd.DataFrame) -> pd.DataFrame:
    """
    Compute the scalar flow duration curve (exceedance probabilities) from two flow duration curves

    Args:
        sim_fdc: simulated flow duration curve
        obs_fdc: observed flow duration curve

    Returns:
        pd.DataFrame with index (exceedance probabilities) and a column of scalars
    """
    scalars_df = pd.DataFrame(np.divide(sim_fdc.values.flatten(), obs_fdc.values.flatten()),
                              columns=['scalars', ],
                              index=sim_fdc.index
                              )
    scalars_df.replace(np.inf, np.nan, inplace=True)
    scalars_df.dropna(inplace=True)
    return scalars_df


def _make_interpolator(x: np.array, y: np.array, extrap: str = 'nearest',
                       fill_value: int or float = None) -> interpolate.interp1d:
    """
    Make an interpolator from two arrays

    Args:
        x: x values
        y: y values
        extrap: method for extrapolation: nearest, const, linear, average, max, min
        fill_value: value to use when extrap='const'

    Returns:
        interpolate.interp1d
    """
    # todo check that flows are not negative and have sufficient variance - even for small variance in SAF
    # if np.max(y) - np.min(y) < 5:
    #     logger.warning('The y data has similar max/min values. You may get unanticipated results.')

    # make interpolator which converts the percentiles to scalars
    if extrap == 'nearest':
        return interpolate.interp1d(x, y, fill_value='extrapolate', kind='nearest')
    elif extrap == 'const':
        if fill_value is None:
            raise ValueError('Must provide the const kwarg when extrap_method="const"')
        return interpolate.interp1d(x, y, fill_value=fill_value, bounds_error=False)
    elif extrap == 'linear':
        return interpolate.interp1d(x, y, fill_value='extrapolate')
    elif extrap == 'average':
        return interpolate.interp1d(x, y, fill_value=np.mean(y), bounds_error=False)
    elif extrap == 'max' or extrap == 'maximum':
        return interpolate.interp1d(x, y, fill_value=np.max(y), bounds_error=False)
    elif extrap == 'min' or extrap == 'minimum':
        return interpolate.interp1d(x, y, fill_value=np.min(y), bounds_error=False)
    else:
        raise ValueError('Invalid extrapolation method provided')


def correct_forecast(forecast_data: pd.DataFrame, simulated_data: pd.DataFrame,
                     observed_data: pd.DataFrame, use_month: int = 0) -> pd.DataFrame:
    """
    Accepts a short term forecast of streamflow, simulated historical flow, and observed flow timeseries and attempts
    to correct biases in the forecasted data

    Args:
        forecast_data: A dataframe with a datetime index and any number of columns of forecasted flow. Compatible with
            forecast_stats, forecast_ensembles, forecast_records
        simulated_data: A dataframe with a datetime index and a single column of streamflow values
        observed_data: A dataframe with a datetime index and a single column of streamflow values
        use_month: Optional: either 0 for correct the forecast based on the first month of the forecast data or -1 if
            you want to correct based on the ending month of the forecast data

    Returns:
        pandas DataFrame with a copy of forecasted data with values updated in each column
    """
    # make a copy of the forecasts which we update and return so the original data is not changed
    forecast_copy = forecast_data.copy()

    # make the flow and probability interpolation functions
    monthly_simulated = simulated_data[simulated_data.index.month == forecast_copy.index[use_month].month].dropna()
    monthly_observed = observed_data[observed_data.index.month == forecast_copy.index[use_month].month].dropna()
    to_prob = _flow_and_probability_mapper(monthly_simulated, to_probability=True, extrapolate=True)
    to_flow = _flow_and_probability_mapper(monthly_observed, to_flow=True, extrapolate=True)

    # for each column of forecast data, make the interpolation function and update the dataframe
    for column in forecast_copy.columns:
        tmp = forecast_copy[column].dropna()
        forecast_copy.update(pd.DataFrame(to_flow(to_prob(tmp.values)), index=tmp.index, columns=[column]))

    return forecast_copy


def statistics_tables(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                      merged_sim_obs: pd.DataFrame = False, merged_cor_obs: pd.DataFrame = False,
                      metrics: list = None) -> str:
    """
    Makes an html table of various statistical metrics for corrected vs observed data alongside the same metrics for
    the simulated vs observed data as a way to see the improvement made by the bias correction. This function uses
    hydrostats.data.merge_data on the 3 inputs. If you have already computed these because you are doing a full
    comparison of bias correction, you can provide them to save time

    Args:
        corrected: A dataframe with a datetime index and a single column of streamflow values
        simulated: A dataframe with a datetime index and a single column of streamflow values
        observed: A dataframe with a datetime index and a single column of streamflow values
        merged_sim_obs: (optional) if you have already computed it, hydrostats.data.merge_data(simulated, observed)
        merged_cor_obs: (optional) if you have already computed it, hydrostats.data.merge_data(corrected, observed)
        metrics: A list of abbreviated statistic names. See the documentation for HydroErr
    """
    if corrected is False and simulated is False and observed is False:
        if merged_sim_obs is not False and merged_cor_obs is not False:
            pass  # if you provided the merged dataframes already, we use those
    else:
        # merge the datasets together
        merged_sim_obs = hd.merge_data(sim_df=simulated, obs_df=observed)
        merged_cor_obs = hd.merge_data(sim_df=corrected, obs_df=observed)

    if metrics is None:
        metrics = ['ME', 'RMSE', 'NRMSE (Mean)', 'MAPE', 'NSE', 'KGE (2009)', 'KGE (2012)']
    # Merge Data
    table1 = hs.make_table(merged_dataframe=merged_sim_obs, metrics=metrics)
    table2 = hs.make_table(merged_dataframe=merged_cor_obs, metrics=metrics)

    table2 = table2.rename(index={'Full Time Series': 'Corrected Full Time Series'})
    table1 = table1.rename(index={'Full Time Series': 'Original Full Time Series'})
    table1 = table1.transpose()
    table2 = table2.transpose()

    table_final = pd.merge(table1, table2, right_index=True, left_index=True)

    return table_final.to_html()


def _flow_and_probability_mapper(monthly_data: pd.DataFrame, to_probability: bool = False,
                                 to_flow: bool = False, extrapolate: bool = False) -> interpolate.interp1d:
    if not to_flow and not to_probability:
        raise ValueError('You need to specify either to_probability or to_flow as True')

    # get maximum value to bound histogram
    max_val = math.ceil(np.max(monthly_data.max()))
    min_val = math.floor(np.min(monthly_data.min()))

    if max_val == min_val:
        warnings.warn('The observational data has the same max and min value. You may get unanticipated results.')
        max_val += .1

    # determine number of histograms bins needed
    number_of_points = len(monthly_data.values)
    number_of_classes = math.ceil(1 + (3.322 * math.log10(number_of_points)))

    # specify the bin width for histogram (in m3/s)
    step_width = (max_val - min_val) / number_of_classes

    # specify histogram bins
    bins = np.arange(-np.min(step_width), max_val + 2 * np.min(step_width), np.min(step_width))

    if bins[0] == 0:
        bins = np.concatenate((-bins[1], bins))
    elif bins[0] > 0:
        bins = np.concatenate((-bins[0], bins))

    # make the histogram
    counts, bin_edges = np.histogram(monthly_data, bins=bins)

    # adjust the bins to be the center
    bin_edges = bin_edges[1:]

    # normalize the histograms
    counts = counts.astype(float) / monthly_data.size

    # calculate the cdfs
    cdf = np.cumsum(counts)

    # interpolated function to convert simulated streamflow to prob
    if to_probability:
        if extrapolate:
            func = interpolate.interp1d(bin_edges, cdf, fill_value='extrapolate')
        else:
            func = interpolate.interp1d(bin_edges, cdf)
        return lambda x: np.clip(func(x), 0, 1)
    # interpolated function to convert simulated prob to observed streamflow
    elif to_flow:
        if extrapolate:
            return interpolate.interp1d(cdf, bin_edges, fill_value='extrapolate')
        return interpolate.interp1d(cdf, bin_edges)



import numpy as np
import pandas as pd
import plotly.graph_objs as go
import scipy.stats

from .format_tools import build_title, timezone_label
from .plotly_helpers import _rperiod_scatters

__all__ = [
    'retrospective',
    'daily_averages',
    'monthly_averages',
    'annual_averages',
    'daily_variance',
    'flow_duration_curve',
]


def retrospective(df: pd.DataFrame, *, rp_df: pd.DataFrame = None, plot_titles: dict = None, ) -> go.Figure:
    """
    Makes the streamflow ensemble data and metadata into a plotly plot

    Args:
        df: the csv response from historic_simulation
        rp_df: the csv response from return_periods
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    dates = df.index.tolist()
    startdate = dates[0]
    enddate = dates[-1]

    plot_data = {
        'x_datetime': dates,
        'y_flow': df.values.flatten(),
        'y_max': df.values.max(),
    }
    if rp_df is not None:
        plot_data.update(rp_df.to_dict(orient='index').items())
        rperiod_scatters = _rperiod_scatters(startdate, enddate, rp_df, plot_data['y_max'], plot_data['y_max'])
    else:
        rperiod_scatters = []

    scatter_plots = [go.Scatter(
        name='Retrospective Simulation',
        x=plot_data['x_datetime'],
        y=plot_data['y_flow'])
    ]
    scatter_plots += rperiod_scatters

    layout = go.Layout(
        title=build_title('Retrospective Streamflow Simulation', plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
        xaxis={
            'title': timezone_label(df.index.tz),
            'range': [startdate, enddate],
            'hoverformat': '%d %b %Y',
            'tickformat': '%Y'
        },
    )
    return go.Figure(scatter_plots, layout=layout)


def daily_averages(dayavg: pd.DataFrame, plot_titles: list = None, plot_type: str = 'plotly') -> go.Figure:
    """
    Makes the daily_averages data and metadata into a plotly plot

    Args:
        dayavg: the csv response from daily_averages
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}
        plot_type: either 'plotly', or 'html' (default plotly)

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if plot_type not in ['plotly_scatters', 'plotly', 'html']:
        raise ValueError('invalid plot_type specified. pick plotly, plotly_scatters, or html')

    scatter_plots = [
        go.Scatter(
            name='Average Daily Flow',
            x=dayavg.index.tolist(),
            y=dayavg.values.flatten(),
            line=dict(color='blue')
        ),
    ]
    if plot_type == 'plotly_scatters':
        return scatter_plots
    layout = go.Layout(
        title=build_title('Daily Average Streamflow (Simulated)', plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
        xaxis={'title': 'Date', 'hoverformat': '%b %d', 'tickformat': '%b'},
    )
    return go.Figure(scatter_plots, layout=layout)


def monthly_averages(monavg: pd.DataFrame, plot_titles: list = None,
                     plot_type: str = 'plotly') -> go.Figure:
    """
    Makes the daily_averages data and metadata into a plotly plot

    Args:
        monavg: the csv response from monthly_averages
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}
        plot_type: either 'plotly', or 'html' (default plotly)

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if plot_type not in ['plotly_scatters', 'plotly', 'html']:
        raise ValueError('invalid plot_type specified. pick plotly, plotly_scatters, or html')

    scatter_plots = [
        go.Scatter(
            name='Average Monthly Flow',
            x=pd.to_datetime(monavg.index, format='%m').strftime('%B'),
            y=monavg.values.flatten(),
            line=dict(color='blue')
        ),
    ]
    if plot_type == 'plotly_scatters':
        return scatter_plots
    layout = go.Layout(
        title=build_title('Monthly Average Streamflow (Simulated)', plot_titles=plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)'},
        xaxis={'title': 'Month'},
    )
    return go.Figure(scatter_plots, layout=layout)


def annual_averages(df: pd.DataFrame, *, plot_titles: list = None, decade_averages: bool = False) -> go.Figure:
    """
    Makes the annual_averages data and metadata into a plotly plot

    Args:
        df: the csv response from annual_averages
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}
        decade_averages: (bool) if True, will plot the average flow for each decade

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    scatter_plots = [
        go.Scatter(
            name='Average Annual Flow',
            x=df.index.tolist(),
            y=df.values.flatten(),
            line=dict(color='blue')
        ),
    ]

    if decade_averages:
        # get a list of decades covered by the data in the index
        first_year = str(int(df.index[0]))[:-1] + '0'
        last_year = str(int(df.index[-1]))[:-1] + '9'
        first_year = int(first_year)
        last_year = int(last_year)
        decades = [decade for decade in range(int(first_year), int(last_year) + 1, 10)]
        for idx, decade in enumerate(decades):
            decade_values = df[np.logical_and(df.index.astype(int) >= decade, df.index.astype(int) < decade + 10)]
            mean_flow = decade_values.values.flatten().mean()
            scatter_plots.append(
                go.Scatter(
                    name=f'{decade}s: {mean_flow:.2f} m<sup>3</sup>/s',
                    x=[decade_values.index[0], decade_values.index[-1]],
                    y=mean_flow * np.ones(2),
                    line=dict(color='red'),
                    hoverinfo='name',
                    legendgroup='decade_averages',
                    legendgrouptitle=dict(text='Decade Averages')
                )
            )

    layout = go.Layout(
        title=build_title('Annual Average Streamflow (Simulated)', plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)'},
        xaxis={'title': 'Year'},
    )
    return go.Figure(scatter_plots, layout=layout)


def flow_duration_curve(df: pd.DataFrame, plot_titles: dict = None, plot_type: str = 'plotly') -> go.Figure:
    """
    Makes the streamflow ensemble data and metadata into a plotly plot

    Args:
        df: the dataframe response from data.retrospective
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}
        plot_type: either 'json', 'plotly', or 'html' (default plotly)

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if plot_type not in ['json', 'plotly_scatters', 'plotly', 'html']:
        raise ValueError('invalid plot_type specified. pick json, plotly, plotly_scatters, or html')

    # process the hist dataframe to create the flow duration curve
    sorted_hist = df.values.flatten()
    sorted_hist.sort()

    # ranks data from smallest to largest
    ranks = len(sorted_hist) - scipy.stats.rankdata(sorted_hist, method='average')

    # calculate probability of each rank
    prob = [(ranks[i] / (len(sorted_hist) + 1)) for i in range(len(sorted_hist))]

    plot_data = {
        'x_probability': prob,
        'y_flow': sorted_hist,
        'y_max': sorted_hist[0],
    }

    if plot_type == 'json':
        return plot_data

    scatter_plots = [
        go.Scatter(
            name='Flow Duration Curve',
            x=plot_data['x_probability'],
            y=plot_data['y_flow'])
    ]
    if plot_type == 'plotly_scatters':
        return scatter_plots
    layout = go.Layout(
        title=build_title('Flow Duration Curve', plot_titles),
        xaxis={'title': 'Exceedence Probability'},
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
    )
    return go.Figure(scatter_plots, layout=layout)


def daily_stats(df: pd.DataFrame, *, plot_titles: dict = None, plot_type: str = 'plotly') -> go.Figure:
    """
    Plots a graph with statistics for each day of year

    Args:
        df: dataframe of values to plot
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}
        plot_type: either 'plotly' (python object, default), 'plotly_scatters', or 'html'

    returns:
        plot of the graph of the low flows
    """

    stats_df = daily_stats(df)

    data = [
        go.Scatter(
            x=stats_df.index.tolist(),
            y=stats_df[column].tolist(),
            name=column
        ) for column in stats_df.columns
    ]

    if plot_type == 'plotly_scatters':
        return data
    layout = go.Layout(
        title=build_title('Daily Average Streamflow (Simulated)', plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
        xaxis={'title': timezone_label(df.index.tz), 'hoverformat': '%b %d', 'tickformat': '%b'},
    )
    return go.Figure(data=data, layout=layout)


def daily_variance(daily_variance: pd.DataFrame, plot_titles: list = None, plot_type: str = 'plotly') -> go.Figure:
    """
    A dataframe of daily variances computed by the geoglows.analysis.compute_daily_variance function

    Args:
      daily_variance: dataframe of values to plot coming from geoglows.analysis.compute_daily_variance
      plot_titles: (dict) Extra info to show on the title of the plot. For example:
        {'River ID': 1234567, 'Drainage Area': '1000km^2'}
      plot_type: either 'plotly' (python object, default), 'plotly_scatters', or 'html'

    returns:
      plot of standard deviation and plot of the graph of the low flows
    """
    data = [
        go.Scatter(
            x=daily_variance['date'],
            y=daily_variance['flow_std'],
            name="Daily Standard Deviation"
        ),
    ]
    if plot_type == 'plotly_scatters':
        return data
    return go.Figure(data=data, layout=go.Layout(build_title('Daily Flow Standard Deviation', plot_titles)))



import os
import warnings

import numpy as np
import pandas as pd
import xarray as xr

from ._constants import get_metadata_table_path, get_sfdc_table_path, get_saber_assign_table_path, get_fdc_table_path, get_sfdc_zarr_uri, get_transformer_table_uri
from ._download_decorators import _forecast, _retrospective, DEFAULT_REST_ENDPOINT, DEFAULT_REST_ENDPOINT_VERSION
from .analyze import (
    daily_averages as calc_daily_averages,
    monthly_averages as calc_monthly_averages,
    annual_averages as calc_annual_averages,
)

__all__ = [
    # forecast products
    'dates',
    'forecast',
    'forecast_stats',
    'forecast_ensembles',
    'forecast_records',

    # retrospective products
    'retrospective',
    'daily_averages',
    'monthly_averages',
    'annual_averages',
    'return_periods',

    # transformers
    'sfdc',
    'assigned_sfdc_curve_id',
    'sfdc_for_river_id',

    'metadata_tables',
    'saber_assign_table',

    'DEFAULT_REST_ENDPOINT',
    'DEFAULT_REST_ENDPOINT_VERSION',
]


# Forecast data and derived products
@_forecast
def dates(**kwargs) -> dict or str:
    """
    Gets a list of available forecast product dates

    Keyword Args:
        data_source (str): location to query for data, either 'rest' or 'aws'. default is aws.

    Returns:
        dict or str

        the csv is a single column with a header of 'available_dates' and 1 row per date, sorted oldest to newest
        The dictionary structure is {'available_dates': ['list', 'of', 'dates', 'YYYYMMDD', 'format']}
    """
    pass


@_forecast
def forecast(*, river_id: int, date: str, format: str, data_source: str,
             **kwargs) -> pd.DataFrame or xr.Dataset:
    """
    Gets the average forecasted flow for a certain river_id on a certain date

    Keyword Args:
        river_id (int): the ID of a stream, should be a 9 digit integer
        date (str): a string specifying the date to request in YYYYMMDD format, returns the latest available if not specified
        format: if data_source=="rest": csv, json, or url, default csv. if data_source=="aws": df or xarray
        data_source (str): location to query for data, either 'rest' or 'aws'. default is aws.

    Returns:
        pd.DataFrame or dict or str
    """
    pass


@_forecast
def forecast_stats(*, river_id: int, date: str, format: str, data_source: str,
                   **kwargs) -> pd.DataFrame or xr.Dataset:
    """
    Retrieves the min, 25%, mean, median, 75%, and max river discharge of the 51 ensembles members for a river_id
    The 52nd higher resolution member is excluded

    Keyword Args:
        river_id (int): the ID of a stream, should be a 9 digit integer
        date (str): a string specifying the date to request in YYYYMMDD format, returns the latest available if not specified
        format (str): if data_source=="rest": csv, json, or url, default csv. if data_source=="aws": df or xarray
        data_source (str): location to query for data, either 'rest' or 'aws'. default is aws.

    Returns:
        pd.DataFrame or dict or str
    """
    pass


@_forecast
def forecast_ensembles(*, river_id: int, date: str, format: str, data_source: str,
                       **kwargs) -> pd.DataFrame or xr.Dataset:
    """
    Retrieves each of 52 time series of forecasted discharge for a river_id on a certain date

    Keyword Args:
        river_id (int): the ID of a stream, should be a 9 digit integer
        date (str): a string specifying the date to request in YYYYMMDD format, returns the latest available if not specified
        format (str): if data_source=="rest": csv, json, or url, default csv. if data_source=="aws": df or xarray
        data_source (str): location to query for data, either 'rest' or 'aws'. default is aws.

    Returns:
        pd.DataFrame or dict or str
    """
    pass


@_forecast
def forecast_records(*, river_id: int, start_date: str, end_date: str, format: str,
                     **kwargs) -> pd.DataFrame or dict or str:
    """
    Retrieves a csv showing the ensemble average forecasted flow for the year from January 1 to the current date

    Keyword Args:
        river_id (int): the ID of a stream, should be a 9 digit integer
        start_date (str): a YYYYMMDD string giving the earliest date this year to include, defaults to 14 days ago.
        end_date (str): a YYYYMMDD string giving the latest date this year to include, defaults to latest available
        format (str): csv, json, or url, default csv.

    Returns:
        pd.DataFrame or dict or str
    """
    pass


# Retrospective simulation and derived products
@_retrospective
def retrospective(river_id: int or list, *, format: str = 'df') -> pd.DataFrame or xr.Dataset:
    """
    Retrieves the retrospective simulation of streamflow for a given river_id from the
    AWS Open Data Program GEOGLOWS V2 S3 bucket

    Args:
        river_id (int): the ID of a stream, should be a 9 digit integer
        format (str): the format to return the data, either 'df' or 'xarray'. default is 'df'

    Returns:
        pd.DataFrame
    """
    pass


def daily_averages(river_id: int or list, **kwargs) -> pd.DataFrame:
    """
    Retrieves daily average streamflow for a given river_id

    Args:
        river_id (int): the ID of a stream, should be a 9 digit integer

    Returns:
        pd.DataFrame
    """
    df = retrospective(river_id, **kwargs)
    return calc_daily_averages(df)


def monthly_averages(river_id: int or list, **kwargs) -> pd.DataFrame:
    """
    Retrieves monthly average streamflow for a given river_id

    Args:
        river_id (int): the ID of a stream, should be a 9 digit integer

    Returns:
        pd.DataFrame
    """
    df = retrospective(river_id, **kwargs)
    return calc_monthly_averages(df)


def annual_averages(river_id: int or list, **kwargs) -> pd.DataFrame:
    """
    Retrieves annual average streamflow for a given river_id

    Args:
        river_id (int): the ID of a stream, should be a 9 digit integer

    Returns:
        pd.DataFrame
    """
    df = retrospective(river_id, **kwargs)
    return calc_annual_averages(df)


def assigned_sfdc_curve_id(river_id: int or list) -> list:
    """
    Retrieves 'asgn_mid' values from the SABER assign table for the given river_id(s).

    Args:
        river_id (int or list): ID(s) of a stream(s), should be a 9-digit integer or a list of such integers.

    Returns:
        list: List of 'asgn_mid' values for given river_id.
    """
    if hasattr(river_id, 'item'):
        river_id = river_id.item()
    if isinstance(river_id, (int, np.integer)):
        river_id = [river_id]
    elif isinstance(river_id, list):
        if not all(isinstance(x, int) for x in river_id):
            raise ValueError("All river_id values must be integers")
    else:
        raise ValueError("river_id must be an integer or a list of integers")
    if hasattr(river_id, 'item'):
        river_id = river_id.item()

    # Read the SABER assign table into a DataFrame
    df = pd.read_parquet(get_transformer_table_uri())
    curve_ids = df.loc[df['river_id'].isin(river_id), 'sfdc_curve_id'].tolist()
    return curve_ids


def sfdc(curve_id: int or list) -> pd.DataFrame:
    """
    Retrieves data from the SFDC table based on 'asgn_mid' values for given river_id.

    Args:
        curve_id (int or list): Single or list of sfdc curve IDs

    Returns:
        pd.DataFrame
    """
    # check that curve_id is a 12 digit integer or a list of such integers
    if isinstance(curve_id, (int, np.integer)):
        assert len(str(curve_id)) == 12, "curve_id must be a 12 digit integer"
    if isinstance(curve_id, list):
        assert all(len(str(x)) == 12 for x in curve_id), "curve_id must be a 12 digit integer"
        assert all(isinstance(x, int) for x in curve_id), "curve_id must be a 12 digit integer"
    ds = xr.open_zarr(get_sfdc_zarr_uri())
    return ds.sel(curve_id=curve_id).to_dataframe().reset_index()


def sfdc_for_river_id(river_id: int or list) -> pd.DataFrame:
    """
    Retrieves data from the SFDC table using 'asgn_mid' values obtained from the SABER assign table for the given 'river_id'.

    Args:
        river_id (int or list): ID(s) of a stream(s).

    Returns:
        pd.DataFrame: Filtered DataFrame from the SFDC table based on 'asgn_mid' values.
    """
    curve_ids = assigned_sfdc_curve_id(river_id)
    filtered_sfdc = sfdc(curve_ids)
    return filtered_sfdc


@_retrospective
def return_periods(river_id: int or list, *, format: str = 'df', method: str = 'gumbel1') -> pd.DataFrame or xr.Dataset:
    """
    Retrieves the return period thresholds based on a specified historic simulation forcing on a certain river_id.

    Args:
        river_id (int): the ID of a stream, should be a 9 digit integer
        format (str): the format to return the data, either 'df' or 'xarray'. default is 'df'
        method (str): the method to use to estimate the return period thresholds. default is 'gumbel1'

    Changelog:
        v1.4.0: adds method parameter for future expansion of multiple return period methods

    Returns:
        pd.DataFrame
    """
    pass


# model config and supplementary data
def metadata_tables(columns: list = None, metadata_table_path: str = None) -> pd.DataFrame:
    """
    Retrieves the master table of rivers metadata and properties as a pandas DataFrame
    Args:
        columns (list): optional subset of columns names to read from the parquet
        metadata_table_path (str): optional path to a local copy of the metadata table

    Returns:
        pd.DataFrame
    """
    if metadata_table_path:
        return pd.read_parquet(metadata_table_path, columns=columns)
    metadata_table_path = get_metadata_table_path()
    if os.path.exists(metadata_table_path):
        return pd.read_parquet(metadata_table_path, columns=columns)
    warn = f"""
    Local copy of geoglows v2 metadata table not found.
    A copy of the table has been cached at {metadata_table_path} which you can move as desired.
    You should set the environment variable PYGEOGLOWS_METADATA_TABLE_PATH or provide the metadata_table_path argument.
    """
    warnings.warn(warn)
    df = pd.read_parquet('http://geoglows-v2.s3-website-us-west-2.amazonaws.com/tables/package-metadata-table.parquet')
    os.makedirs(os.path.dirname(metadata_table_path), exist_ok=True)
    df.to_parquet(metadata_table_path)
    return df[columns] if columns else df


def saber_assign_table(columns: list = None, saber_table_path: str = None) -> pd.DataFrame:
    """
    Retrieves the SABER assign table as a pandas DataFrame

    Args:
        columns (list): optional subset of column names to read from the parquet
        saber_table_path (str): optional path to a local copy of the SABER assign table

    Returns:
        pd.DataFrame
    """
    if saber_table_path:
        return pd.read_parquet(saber_table_path, columns=columns)
    saber_table_path = get_saber_assign_table_path()
    if os.path.exists(saber_table_path):
        return pd.read_parquet(saber_table_path, columns=columns)

    warn = f"""
    Local copy of SABER assign table not found.
    A copy of the table has been cached at {saber_table_path} which you can move as desired.
    You should set the environment variable PYGEOGLOWS_SABER_ASSIGN_TABLE_PATH or provide the saber_table_path argument.
    """
    warnings.warn(warn)

    # Assuming the S3 URL for the SABER assign table. You may need to adjust this URL.
    df = pd.read_parquet('http://geoglows-v2.s3-website-us-west-2.amazonaws.com/tables/saber-assign-table.parquet')

    os.makedirs(os.path.dirname(saber_table_path), exist_ok=True)
    df.to_parquet(saber_table_path)

    return df[columns] if columns else df



from .plots import (
    forecast,
    forecast_stats,
    forecast_ensembles,
    forecast_records,
    retrospective,
    daily_averages,
    monthly_averages,
    annual_averages,
    flow_duration_curve,
    corrected_retrospective,
    corrected_month_average,
    corrected_day_average,
    corrected_scatterplots,

    plotly_figure_to_html_plot as plotly_figure_to_html,
)

__all__ = [
    'forecast',
    'forecast_stats',
    'forecast_ensembles',
    'forecast_records',
    'retrospective',
    'daily_averages',
    'monthly_averages',
    'annual_averages',
    'flow_duration_curve',
    'corrected_retrospective',
    'corrected_month_average',
    'corrected_day_average',
    'corrected_scatterplots',
]



import json
import warnings
from io import StringIO

import pandas as pd
import requests

ENDPOINT = 'https://geoglows.ecmwf.int/api/'

DEPRECATIONWARNING = """
The streamflow module is deprecated and will be removed early 2025 when GEOGLOWS Model V1 is removed. These functions 
will no longer be updated and they will not work with the latest model and datasets. Please upgrade to GEOGLOWS Model V2 
and update your code to use the geoglows.data module's analogous functions. Visit https://data.geoglows.org for more 
information and tutorials to help you transition.
"""

__all__ = ['forecast_stats', 'forecast_ensembles', 'forecast_warnings', 'forecast_records', 'historic_simulation',
           'daily_averages', 'monthly_averages', 'return_periods', 'available_dates', ]


# FUNCTIONS THAT CALL THE GLOBAL STREAMFLOW PREDICTION API
def forecast_stats(reach_id: int, return_format: str = 'csv', forecast_date: str = None,
                   endpoint: str = ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves statistics that summarize the ensemble streamflow forecast for a certain reach_id

    Args:
        reach_id: the ID of a stream
        return_format: 'csv', 'json', 'waterml', 'url'
        forecast_date: a string specifying the date to request in YYYYMMDD format
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.forecast_stats(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'ForecastStats/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}'
    params = {'reach_id': reach_id, 'return_format': return_format}
    if forecast_date is not None:
        params["date"] = forecast_date
    # return the requested data
    return _make_request(endpoint, method, params, return_format, s)


def forecast_ensembles(reach_id: int, return_format: str = 'csv', forecast_date: str = None,
                       endpoint: str = ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves each ensemble from the most recent streamflow forecast for a certain reach_id

    Args:
        reach_id: the ID of a stream
        return_format: 'csv', 'json', 'waterml', 'url'
        forecast_date: a string specifying the date to request in YYYYMMDD format
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.forecast_ensembles(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'ForecastEnsembles/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}'

    params = {'reach_id': reach_id, 'return_format': return_format}
    if forecast_date is not None:
        params["date"] = forecast_date

    # return the requested data
    return _make_request(endpoint, method, params, return_format, s)


def forecast_warnings(region: str = 'all', return_format='csv',
                      endpoint=ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves a csv listing streams likely to experience a return period level flow during the forecast period.

    Args:
        region: the name of a region as shown in the available_regions request
        return_format: 'csv', 'json', 'waterml', 'request', 'url'
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.forecast_warnings('australia-geoglows')
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'ForecastWarnings/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return endpoint + method + f'?region={region}'

    # return the requested data
    return _make_request(endpoint, method, {'region': region, 'return_format': return_format}, return_format, s)


def forecast_records(reach_id: int, start_date: str = None, end_date: str = None, return_format='csv',
                     endpoint=ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves a csv showing the ensemble average forecasted flow for the year from January 1 to the current date

    Args:
        reach_id: the ID of a stream
        return_format: 'csv', 'json', 'waterml', 'url'
        start_date: a string specifying the earliest date to request in YYYYMMDD format
        end_date: a string specifying the latest date to request in YYYYMMDD format
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.forecast_warnings('australia-geoglows')
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'ForecastRecords/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}'

    params = {'reach_id': reach_id, 'return_format': return_format}
    if start_date is not None:
        params["start_date"] = start_date
    if end_date is not None:
        params["end_date"] = end_date

    # return the requested data
    return _make_request(endpoint, method, params, return_format, s)


def historic_simulation(reach_id: int, return_format='csv', forcing='era_5',
                        endpoint=ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves a historical streamflow simulation derived from a specified forcing for a certain reach_id

    Args:
        reach_id: the ID of a stream
        return_format: 'csv', 'json', 'waterml', 'url'
        forcing: the runoff dataset used to drive the historic simulation (era_interim or era_5)
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.historic_simulation(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'HistoricSimulation/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}&forcing={forcing}'

    # return the requested data
    params = {'reach_id': reach_id, 'forcing': forcing, 'return_format': return_format}
    return _make_request(endpoint, method, params, return_format, s)


def daily_averages(reach_id: int, return_format='csv', forcing='era_5',
                   endpoint=ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves the average flow for every day of the year at a certain reach_id.

    Args:
        reach_id: the ID of a stream
        return_format: 'csv', 'json', 'waterml', 'url'
        forcing: the runoff dataset used to drive the historic simulation (era_interim or era_5)
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.seasonal_average(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'DailyAverages/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}&forcing={forcing}'

    # return the requested data
    params = {'reach_id': reach_id, 'forcing': forcing, 'return_format': return_format}
    return _make_request(endpoint, method, params, return_format, s)


def monthly_averages(reach_id: int, return_format='csv', forcing='era_5',
                     endpoint=ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves the average flow for each month at a certain reach_id.

    Args:
        reach_id: the ID of a stream
        forcing: the runoff dataset used to drive the historic simulation (era_interim or era_5)
        return_format: 'csv', 'json', 'waterml', 'url'
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.seasonal_average(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'MonthlyAverages/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}&forcing={forcing}'

    # return the requested data
    params = {'reach_id': reach_id, 'forcing': forcing, 'return_format': return_format}
    return _make_request(endpoint, method, params, return_format, s)


def return_periods(reach_id: int, return_format='csv', forcing='era_5',
                   endpoint=ENDPOINT, s: requests.Session = False) -> pd.DataFrame:
    """
    Retrieves the return period thresholds based on a specified historic simulation forcing on a certain reach_id.

    Args:
        reach_id: the ID of a stream
        forcing: the runoff dataset used to drive the historic simulation (era_interim or era_5)
        return_format: 'csv', 'json', 'waterml', 'url'
        endpoint: the endpoint of an api instance
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='csv' returns a pd.DataFrame()
        - return_format='json' returns a json
        - return_format='waterml' returns a waterml string
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.return_periods(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'ReturnPeriods/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return f'{endpoint}{method}?reach_id={reach_id}&forcing={forcing}'

    # return the requested data
    params = {'reach_id': reach_id, 'forcing': forcing, 'return_format': return_format}

    return _make_request(endpoint, method, params, return_format, s)


def available_data(endpoint: str = ENDPOINT, return_format='json', s: requests.Session = False) -> dict or str:
    """
    Returns a dictionary with a key for each available_regions containing the available_dates for that region

    Args:
        endpoint: the endpoint of an api instance
        return_format: 'json' or 'url'
        s: requests.Session instance connected to the api's root url

    Returns:
        dict

    Example:
        .. code-block:: python

            data = streamflow.rst.available_data()

    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'AvailableData/'

    # if you only wanted the url, quit here
    if return_format == 'url':
        return endpoint + method

    # return the requested data
    return _make_request(endpoint, method, {}, return_format, s)


def available_dates(reach_id: int = None, region: str = None, return_format: str = 'json',
                    endpoint: str = ENDPOINT, s: requests.Session = False) -> dict or str:
    """
    Retrieves the list of dates of stored streamflow forecasts. You need to specify either a reach_id or a region.

    Args:
        reach_id: the ID of a stream
        region: the name of a hydrologic region used in the model
        endpoint: the endpoint of an api instance
        return_format: 'json' or 'url'
        s: requests.Session instance connected to the api's root url

    Return Format:
        - return_format='json' *(default)* returns {'available_dates': ['list_of_dates']}
        - return_format='url' returns a url string for using in a request or web browser

    Example:
        .. code-block:: python

            data = streamflow.rst.available_dates(12341234)
    """
    warnings.warn(DEPRECATIONWARNING, DeprecationWarning, stacklevel=2)
    method = 'AvailableDates/'

    # you need a region for the api call, so the user needs to provide one or a valid reach_id to get it from
    params = {'region': 'africa-geoglows'}
    # if you only wanted the url, quit here
    if return_format == 'url':
        return endpoint + method

    # return the requested data
    return _make_request(endpoint, method, params, return_format, s)


# API AUXILIARY FUNCTION
def _make_request(endpoint: str, method: str, params: dict, return_format: str, s: requests.Session = False):
    if return_format == 'request':
        params['return_format'] = 'csv'

    # request the data from the API
    if s:
        data = s.get(endpoint + method, params=params)
    else:
        data = requests.get(endpoint + method, params=params)
    if data.status_code != 200:
        raise RuntimeError('Recieved an error from the Streamflow REST API: ' + data.text)

    # process the response from the API as appropriate to make the corresponding python object
    if return_format == 'csv':
        tmp = pd.read_csv(StringIO(data.text), index_col=0)
        if 'z' in tmp.columns:
            del tmp['z']
        if method in ('ForecastWarnings/', 'ReturnPeriods/', 'DailyAverages/', 'MonthlyAverages/'):
            return tmp
        if method == 'SeasonalAverage/':
            tmp.index = pd.to_datetime(tmp.index + 1, format='%j').strftime('%b %d')
            return tmp
        tmp.index = pd.to_datetime(tmp.index)
        return tmp
    elif return_format == 'json':
        return json.loads(data.text)
    elif return_format == 'waterml':
        return data.text
    else:
        raise ValueError(f'Unsupported return format requested: {return_format}')



import os

ODP_CORE_S3_BUCKET_URI = 's3://geoglows-v2'
ODP_FORECAST_S3_BUCKET_URI = 's3://geoglows-v2-forecasts'
ODP_RETROSPECTIVE_S3_BUCKET_URI = 's3://geoglows-v2-retrospective'
ODP_S3_BUCKET_REGION = 'us-west-2'

METADATA_TABLE_PATH = os.getenv(
    'PYGEOGLOWS_METADATA_TABLE_PATH',
    os.path.join(os.path.dirname(__file__), 'data', 'metadata-tables.parquet')
)

SABER_ASSIGN_TABLE_PATH = os.getenv(
    'PYGEOGLOWS_SABER_ASSIGN_TABLE_PATH',
    os.path.join(os.path.dirname(__file__), 'data', 'saber-assign-table.parquet'))

SFDC_TABLE_PATH = os.getenv(
    'PYGEOGLOWS_SFDC_TABLE_PATH',
    os.path.join(os.path.dirname(__file__), 'data', 'sfdc-removing-zero.parquet'))

FDC_TABLE_PATH = os.getenv(
    'PYGEOGLOWS_FDC_TABLE_PATH',
    os.path.join(os.path.dirname(__file__), 'data', 'simulated_monthly.parquet'))


TRANSFORMER_TABLE_ENV_KEY = 'PYGEOGLOWS_TRANSFORMER_TABLE_URI'
SFDC_ZARR_ENV_KEY = 'PYGEOGLOWS_SFDC_ZARR_URI'


def get_sfdc_zarr_uri() -> str:
    return os.getenv(SFDC_ZARR_ENV_KEY, 's3://rfs-v2/transformers/sfdc.zarr')


def set_sfdc_zarr_uri(uri: str) -> None:
    os.environ[SFDC_ZARR_ENV_KEY] = uri
    return


def get_transformer_table_uri() -> str:
    return os.getenv(TRANSFORMER_TABLE_ENV_KEY, 's3://rfs-v2/transformers/transformer_table.parquet')


def set_transformer_table_uri(uri: str) -> None:
    os.environ[TRANSFORMER_TABLE_ENV_KEY] = uri
    return


def get_metadata_table_path() -> str:
    return os.getenv('PYGEOGLOWS_METADATA_TABLE_PATH', METADATA_TABLE_PATH)


def set_metadata_table_path(path: str) -> str:
    global METADATA_TABLE_PATH
    METADATA_TABLE_PATH = path
    return METADATA_TABLE_PATH


def get_saber_assign_table_path() -> str:
    return SABER_ASSIGN_TABLE_PATH


def set_saber_assign_table_path(path: str) -> str:
    global SABER_ASSIGN_TABLE_PATH
    SABER_ASSIGN_TABLE_PATH = path
    return SABER_ASSIGN_TABLE_PATH


def get_sfdc_table_path() -> str:
    return SFDC_TABLE_PATH


def set_sfdc_table_path(path: str) -> str:
    global SFDC_TABLE_PATH
    SFDC_TABLE_PATH = path
    return SFDC_TABLE_PATH


def get_fdc_table_path() -> str:
    return FDC_TABLE_PATH


def set_fdc_table_path(path: str) -> str:
    global FDC_TABLE_PATH
    FDC_TABLE_PATH = path
    return FDC_TABLE_PATH
